use clap::ValueEnum;
use log::{debug, info, warn};
use serde::Serialize;
use std::net::{IpAddr, UdpSocket};
use tailstun::{StunMessage, TxId};

#[derive(Debug, Clone, ValueEnum)]
enum OutputFormat {
    Text,
    Json,
    Yaml,
}

impl OutputFormat {
    fn format_stun(&self, msg: &StunMessage) -> String {
        match self {
            OutputFormat::Text => format!("{}", msg),
            OutputFormat::Json => serde_json::to_string_pretty(msg).unwrap(),
            OutputFormat::Yaml => serde_yaml::to_string(msg).unwrap(),
        }
    }
    fn format_address(&self, (addr, port): &(IpAddr, u16)) -> String {
        match self {
            OutputFormat::Text => format!("{}", addr),
            OutputFormat::Json => serde_json::to_string_pretty(addr).unwrap(),
            OutputFormat::Yaml => serde_yaml::to_string(addr).unwrap(),
        }
    }
}

#[derive(clap::Parser)]
#[command(version, about, long_about = None)]
#[command(about = "Test a Tailscale derp node's stun service")]
struct Cli {
    host: String,
    #[clap(short, long, default_value = "3478")]
    port: u16,
    #[clap(short, long, default_value = "text")]
    format: OutputFormat,
    #[clap(
        short,
        long,
        default_value = "false",
        help = "Only output the first mapped address"
    )]
    address_only: bool,
    #[command(flatten)]
    verbose: clap_verbosity_flag::Verbosity,
}

fn main() {
    let cli = <Cli as clap::Parser>::parse();
    env_logger::Builder::new()
        .filter_level(cli.verbose.log_level_filter())
        .init();

    let dest = cli.host + ":" + &cli.port.to_string();
    let socket = UdpSocket::bind("[::]:0").expect("Unable to bind a UDP socket");
    socket
        .connect(dest)
        .expect("Unable to connect to the destination");

    info!(
        "Connected to {:?} from {:?}",
        socket.peer_addr(),
        socket.local_addr()
    );

    let req = TxId::new().make_request();

    debug!("Sending request {:?}", &req);

    socket.send(&req).expect("Unable to send request");

    let mut buf = [0u8; 1500];

    if let Ok(received) = socket.recv(&mut buf) {
        debug!("Received response: {:?}", &buf[..received]);

        let msg = StunMessage::parse(&buf).unwrap();
        info!("Parsed message from {}:", socket.peer_addr().unwrap());
        if cli.address_only {
            if let Some(addr) = msg.attributes.mapped_address() {
                println!("{}", cli.format.format_address(addr));
            } else {
                std::process::exit(1);
            }
        } else {
            println!("{}", cli.format.format_stun(&msg));
        }
    } else if let Err(e) = socket.recv(&mut buf) {
        println!("recv function failed: {e:?}");
    }
}
